from importlib.metadata import version


def write_header(path):
    """Write the header for the new Zacros input file.

    Parameters
    ----------
    path: str
        Path of the input file
    """
    with open(path, 'w') as infile:
        infile.write('############################################################################\n')
        infile.write(f"#  Zacros input file generated by ZacrosTools (version {version('zacrostools')+')':<18}  #\n")
        infile.write('#  https://github.com/hprats/ZacrosTools                                   #\n')
        infile.write('#                                                                          #\n')
        infile.write('#  Hector Prats, PhD                                                       #\n')
        infile.write('############################################################################\n\n')
